from . import AWSService, paginateable

class APIGateway(AWSService):
    __servicename__ = 'apigateway'

    @paginateable('get_rest_apis','items','position', ['position','limit'])
    def get_rest_apis(self, position = None, limit = None):
        return self.client.get_rest_apis(**self.get_request_params(locals()))
    
    @paginateable("get_resources", "items", "position", ["position", "limit"])
    def get_resources(self, restApiId, embed = ["methods"], position = None, limit = None):
        return self.client.get_resources(**self.get_request_params(locals()))
    
    @paginateable("get_domain_names", "items", "position", ["position", "limit"])
    def get_domain_names(self, resourceOwner = None, position = None, limit = None):
        return self.client.get_domain_names(**self.get_request_params(locals()))
    
    @paginateable("get_base_path_mappings", "items", "position", ["position", "limit"])
    def get_base_path_mappings(self, domainName, domainNameId = None, position = None, limit = None):
        return self.client.get_base_path_mappings(**self.get_request_params(locals()))
    
    def get_integration(self, restApiId:str, resourceId:str, httpMethod:str):
        return self.client.get_integration(**self.get_request_params(locals()))
    
    def get_method(self, restApiId:str, resourceId:str, httpMethod:str):
        return self.client.get_method(**self.get_request_params(locals()))


from . import AWSService, paginateable


class IdentityStore(AWSService):
    __servicename__ = 'identitystore'

    @paginateable('list_users', 'Users',"NextToken",["NextToken", "MaxResults"])
    def list_users(self, IdentityStoreId, Filters:list = None, MaxResults = None, NextToken = None):
        return self.client.list_users(**self.get_request_params(locals()))



from . import AWSService, paginateable
# As "lambda" is a reserved word in python, the file name changed to "lambda_aws"

class Lambda(AWSService):
    __servicename__ = 'lambda'

    @paginateable("list_functions", "Functions", "NextMarker", ["Marker", "MaxItems"])
    def list_functions(self, MasterRegion = None, FunctionVersion = None, Marker = None, MaxItems:int = None):
        return self.client.list_functions(**self.get_request_params(locals()))
    
    @paginateable("list_event_source_mappings", "EventSourceMappings", "NextMarker", ["Marker", "MaxItems"])
    def list_event_source_mappings(self, EventSourceArn:str = None, FunctionName:str = None, Marker = None, MaxItems:int = None):
        return self.client.list_event_source_mappings(**self.get_request_params(locals()))



from abc import ABC
from functools import wraps
import inspect
from boto3.session import Session
import boto3
from ..utils.common import remove_none_values

_available_services = Session().get_available_services()

class AWSService(ABC):
    __servicename__ = None

    def __init__(self, session:Session = None):
        sn = self.__servicename__

        if sn not in _available_services:
            raise Exception(f"Service {sn} is not available")
        
        self.client = session.client(sn) if session is not None else boto3.client(sn)

    def __getattr__(self, name):
        if hasattr(self.client, name):
            return getattr(self.client, name)
        else:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")


    def _get_all_with_callback(self, get_list_function, result_token_name, next_token_name, callback_function, *args, next_token = None):
        results = []
        next_token = next_token
        has_next = True
        
        while has_next:
            response = get_list_function(*args, next_token)
            result = response.get(result_token_name)

            if callback_function:
                callback_function(result)
            else:
                if result is not None:
                    results.extend(result)
            next_token = response.get(next_token_name)
            has_next = next_token_name in response

        if not callback_function:
            return results
    
    def _get_all(self, get_list_function, result_token_name, next_token_name, *args):
        return self._get_all_with_callback(get_list_function, result_token_name, next_token_name, None, *args)
    
    def get_request_params(self, locals):
        request_params = remove_none_values({k: v for k, v in locals.items() if k != 'self'})
        return request_params

class CustomPaginationIterator:
    def __init__(self, initial_result, iterator, result_token):
        self._initial_result = initial_result
        self._iterator = iterator
        self._result_token = result_token
    
    def __iter__(self):
        yield self._initial_result.get(self._result_token, [])
        for i in self._iterator:
            yield i.get(self._result_token, [])

class ResultWithPagination:
    def __init__(self, initial_result, client, operation_name, result_token, next_token_key, kwargs):
        self._initial_result = initial_result
        self._paginator = client.get_paginator(operation_name)
        self._result_token = result_token
        self._kwargs = kwargs
        self._pagination_start_token = None if initial_result is None else initial_result.get(next_token_key)

    def __getattr__(self, name):
        return getattr(self._initial_result, name)
    
    def paginate(self, pagination_config = None):
        if self._pagination_start_token is None:
            result_items = self._initial_result.get(self._result_token)
            return [result_items] if result_items is not None else []
    
        kwargs = self._kwargs or {}
        kwargs["PaginationConfig"] =  pagination_config or {
            # 'PageSize': 50
        }
        kwargs["PaginationConfig"]["StartingToken"] = self._pagination_start_token

        return CustomPaginationIterator(self._initial_result, self._paginator.paginate(**remove_none_values(kwargs)), self._result_token)

def paginateable(operation_name, result_token, next_token_key, ignore_arguments = []):
    def decorator(func):
        arg_names = list(inspect.signature(func).parameters.keys())[1:] # ignore the self
        @wraps(func)
        def wrapper(*args, **kwargs):
            args_dict = {}

            for i, arg in enumerate(args[1:]):
                args_dict[arg_names[i]] = arg
            args_dict.update(kwargs)

            for ia in ignore_arguments:
                args_dict.pop(ia, None)
                
            result = func(*args, **kwargs)
            return ResultWithPagination(result, args[0].client, operation_name, result_token, next_token_key, args_dict)
        return wrapper
    return decorator


from typing import Union
from . import AWSService, paginateable


class S3(AWSService):
    __servicename__ = 's3'

    @paginateable("list_objects_v2", "Contents", "ContinuationToken", ["ContinuationToken", "MaxKeys"])
    def list_objects_v2(self, Bucket, Delimiter:str = None, EncodingType:str = None, MaxKeys:int = None, Prefix:str = None, ContinuationToken:str = None, FetchOwner:bool = None, StartAfter:str = None, RequestPayer:str = None, ExpectedBucketOwner:str = None, OptionalObjectAttributes:list = None):
        return self.client.list_objects_v2(**self.get_request_params(locals()))

    def get_object(self, s3_path:str) -> bytes:
        bucket_name, key_name = get_bucket_and_key(s3_path)
        response = self.client.get_object(Bucket=bucket_name, Key=key_name)
        return response["Body"].read()
    
    def put_object(self, s3_path:str, body:Union[bytes,str]):
        bucket_name, key_name = get_bucket_and_key(s3_path)
        if isinstance(body,str):
            body = body.encode('utf-8')
        self.client.put_object(Bucket=bucket_name, Key=key_name, Body=body)

def get_bucket_and_key(s3_path:str):
    parts = s3_path.removeprefix('/').split('/',1)
    bucket_name = parts[0]
    key_name = parts[1] if len(parts) > 1 else ""

    return bucket_name, key_name


from . import AWSService, paginateable


class APIGatewayV2(AWSService):
    __servicename__ = 'apigatewayv2'

    @paginateable('get_apis', 'Items', 'NextToken', ['NextToken', "MaxResults"])
    def get_apis(self, NextToken = None, MaxResults = None):
        return self.client.get_apis(**self.get_request_params(locals()))



from . import AWSService, paginateable


class EventBridge(AWSService):
    __servicename__ = 'events'

    @paginateable("list_rules", "Rules", "NextToken", ["NextToken", "Limit"])
    def list_rules(self, NamePrefix = None, EventBusName = None, NextToken = None, Limit = None):
        return self.client.list_rules(**self.get_request_params(locals()))
    
    @paginateable("list_targets_by_rule", "Targets", "NextToken", ["NextToken", "Limit"])
    def list_targets_by_rule(self, Rule, EventBusName = None, NextToken = None, Limit = None):
        return self.client.list_targets_by_rule(**self.get_request_params(locals()))




from . import AWSService, paginateable


class ResourceGroupsTaggingAPI(AWSService):
    __servicename__ = 'resourcegroupstaggingapi'

    @paginateable("get_resources", "ResourceTagMappingList", "PaginationToken", ["PaginationToken", "ResourcesPerPage"])
    def get_resources(self, PaginationToken = None, TagFilters:list = None, ResourcesPerPage:int = None, TagsPerPage:int = None, ResourceTypeFilters:list = None, IncludeComplianceDetails:bool = None, ExcludeCompliantResources:bool = None, ResourceARNList:list = None):
        return self.client.get_resources(**self.get_request_params(locals()))



from . import AWSService, paginateable


class CloudFormation(AWSService):
    __servicename__ = 'cloudformation'

    @paginateable("list_stacks", "StackSummaries", "NextToken", ["NextToken"])
    def list_stacks(self, NextToken = None):
        return self.client.list_stacks(**self.get_request_params(locals()))
    
    @paginateable("list_stack_resources", "StackResourceSummaries", "NextToken", ["NextToken"])
    def list_stack_resources(self, StackName, NextToken = None):
        return self.client.list_stack_resources(**self.get_request_params(locals()))



from . import AWSService, paginateable
from ..utils.common import remove_none_values

class IoT(AWSService):
    __servicename__ = 'iot'

    def describe_thing(self, thingName):
        return self.client.describe_thing(thingName = thingName)
    
    @paginateable("list_thing_principals", "principals", "nextToken", ["nextToken", "maxResults"])
    def list_thing_principals(self, thingName, nextToken = None, maxResults = None):
        return self.client.list_thing_principals(**self.get_request_params(locals()))

    def describe_certificate(self, certificateId):
        return self.client.describe_certificate(certificateId = certificateId)
    
    #TODO: remove
    def get_principals_by_thing(self, thing_name, next_token = None):
        request_params = remove_none_values({
            'thingName':thing_name,
            'nextToken':next_token
        })
        
        return self.client.list_thing_principals(**request_params)
    
    #TODO: remove
    def get_certificate_describe(self, certificate_id):
        return self.client.describe_certificate(certificateId=certificate_id)



from . import AWSService, paginateable


class CloudWatch(AWSService):
    __servicename__ = 'cloudwatch'

    @paginateable("list_metrics", "Metrics", "NextToken", ["NextToken"])
    def list_metrics(self, Namespace:str = None, MetricName:str = None, Dimensions:list = None, RecentlyActive:str = None, IncludeLinkedAccounts:bool = None, OwningAccount:str = None, NextToken = None):
        return self.client.list_metrics(**self.get_request_params(locals()))
